import logging # :( i dont do module level imports
logger = logging.getLogger(__name__)
del logging


class Engine:
    from .rules import Rule
    from typing import Iterable, Literal
    class DeanonPrefix(str): ...
    from pyoxigraph import Store
    from inspect import signature
    from .canon import quads
    def __init__(self,
        rules: Iterable[Rule] = [], *,
        db: Store=Store(),
            MAX_NCYCLES: int=99,
        # safe settings to avoid inf cycling
        # but reduces performance
        derand: Literal['canonicalize'] | DeanonPrefix | Literal[False]  = signature(quads.deanon).parameters['uri'].default,
        # typically expecting the engine to be used in a stand-alone program
        # so it helps to have log_print.
            log: bool=True, log_print: bool=False,
            debug: bool = False,
        ) -> None:
        self.rules = list(rules)
        self.db = db

        self.MAX_NCYCLES = MAX_NCYCLES

        # derand --> (deanon, deanon_uri, canon)
        canon = 'canonicalize'
        if derand not in {canon, False}:
            assert(isinstance(derand, str))
        
        if (derand == canon) or (isinstance(derand, str)):
            self.canon = True
        else:
            self.canon = False
        
        if derand == canon:
            self.deanon = False
        elif derand == False:
            self.deanon = False
        else:
            assert(isinstance(derand, str))
            self.deanon = True
            self.deanon_uri = derand

        self.i = 0
        
        # logging
        if log:
            from collections import namedtuple
            from types import SimpleNamespace as NS
            self.logging = NS(
                print = log_print,
                log = [],
                state = namedtuple('state', ['cycle', 'stored', 'rule', 'new', 'time'] ))
        self.debug = debug

    # TODO: make a method for applying one rule
    def run1(self) -> Store:
        if hasattr(self, 'logging'):
            if self.logging.print:
                line = '-'*10
                logger.info(f"CYCLE {self.i} {line}")

        def process(qs):
            _ = qs
            if self.canon:
                # have to keep it per rule firing. bc too much data potentially.
                from .data import quads
                _ = quads(_)
                from .canon import quads
                _ = quads(_)
                if self.deanon:
                    _ = quads.deanon(_, uri=self.deanon_uri)
            yield from _

        from .db import ingest
        for r in self.rules: # TODO: could be parallelized:
            # before
            if hasattr(self, 'logging'):
                if self.logging.print:
                    logger.info(f"  {repr(r)}")
                from time import monotonic
                start_time = monotonic()
                nquads = len(self.db)
            # do
            _ = r(self.db)
            _ = process(_)
            if self.debug:
                if self.logging.print:
                    _ = tuple(_)
                    for q in _: logger.debug(f"{q}")
            ingest(self.db, _, flush=True)
            # after
            if hasattr(self, 'logging'):
                self.logging.log.append(self.logging.state(
                    cycle=self.i,
                    stored=nquads,
                    rule=r,
                    new=len(self.db)-nquads,
                    time=float('{0:.2f}'.format(monotonic()-start_time))))
                if self.logging.print:
                    s = self.logging.log[-1]
                    width = 6
                    logger.info('   '+f"{s.new:<{width}} quads in {s.time}s"  )
                    del s
            # so if a rule returns a string,
            # it /could/ go in fast in the case of no processing (canon/deanon)
            del _
        self.i += 1
        self.db.flush()
        self.db.optimize()
        if hasattr(self, 'logging'):
            if self.logging.print:
                logger.info(f'db has {len(self.db)} quads')
        return self.db

    def stop(self) -> bool:
        if self.MAX_NCYCLES <= 0:
           return False
        # could put validations here
        if len(self.db) == len(self.run1()):
            return True
        else:
            return False
    
    def __iter__(self) -> Iterable[Store]:
        while (not self.stop()):
            if self.i >= self.MAX_NCYCLES:
                from warnings import warn
                warn('reached max cycles')
                break
            yield self.db
        else: # for case when nothing needs to happen
            yield self.db

    def run(self) -> Store:
        for _ in self: continue
        return self.db
    __call__ = run




from typing import Iterable
from pyoxigraph import Quad, Store
Ingestable = bytes|str|Iterable[Quad]
def ingest(s: Store, d:  Ingestable, *, flush=True):
    if isinstance(d, (str,bytes)):
        # assume ttl
        from pyoxigraph import RdfFormat
        s.bulk_load(d, format=RdfFormat.TURTLE)
    else:
        # assume iterable[quand]
        s.bulk_extend(d)
    
    if flush: s.flush()
    return s




class _reification:
    from typing import Iterable
    from pyoxigraph import Triple, NamedNode
    class terms:
        prefix = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        terms = {'Statement', 'type',
                 'subject', 'predicate', 'object'}
        def __init__(self):
            for t in self.terms:
                setattr(self, t, self.nn(self.prefix, t))
        @staticmethod
        def nn(prefix, term):
            from pyoxigraph import NamedNode
            return NamedNode(f"{prefix}{term}")
    terms = terms()
    def standard(self, str: Iterable[Triple]) -> Iterable[Triple]:
        T = self.Triple
        trm = self.terms
        for t in str:
            if not isinstance(t.subject, T):
                yield t
            else: # nested data triple
                ndt = t.subject
                assert(not  isinstance(t.object,  T))
                #                                  make positive...
                id = self.terms.nn("urn:meta:id:", abs(hash(t))) 
                # ...probably wont make use of the number (specifically)
                yield T(id, trm.type,       trm.Statement)
                yield T(id, trm.subject,    ndt.subject)
                yield T(id, trm.predicate,  ndt.predicate)
                yield T(id, trm.object,     ndt.object)
                # meta
                yield T(id, t.predicate,    t.object)
    # :man :hasSpouse :woman .
    # :id1 rdf:type rdf:Statement ;
        # rdf:subject :man ;
        # rdf:predicate :hasSpouse ;
        # rdf:object :woman ;
        # :startDate "2020-02-11"^^xsd:date .
    # <<:man :hasSpouse :woman>> :startDate "2020-02-11"^^xsd:date .
    class query:
        # https://www.ontotext.com/knowledgehub/fundamentals/what-is-rdf-star/
        class parts:
            std = """
            ?reification a rdf:Statement .
            ?reification rdf:subject ?subject .
            ?reification rdf:predicate ?predicate .
            ?reification rdf:object ?object .
            ?reification ?p ?o .
            """
        std2str = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        CONSTRUCT {
            <<?subject ?predicate ?object>> ?p ?o .
        } WHERE {
            parts.std
            FILTER (?p NOT IN (rdf:subject, rdf:predicate, rdf:object) &&
            (?p != rdf:type && ?object != rdf:Statement))
        }
        """
        std2str = std2str.replace('parts.std', parts.std)
        delstd = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        DELETE where { parts.std }
        """
        delstd = delstd.replace('parts.std', parts.std)
    def star(self, std: Iterable[Triple]) -> Iterable[Triple]:
        from pyoxigraph import Store, Quad
        s = Store()
        s.bulk_extend(Quad(*t) for t in std)
        yield from s.query( self.query.std2str)
        s.update(           self.query.delstd)   # delete std
        #                             the 'regular' ones
        yield from s.query("construct {?s ?p ?o} where {?s ?p ?o.}")


reification = _reification()


from .db import Ingestable
from typing import Iterable
from pyoxigraph import Quad
def quads(i: Ingestable) -> Iterable[Quad]:
    from .db import ingest, Store
    yield from ingest(Store(), i)




# arg parsing
from sys import argv
if not len(argv)==2:
    raise ValueError('error: give spec file')
spec = argv[1]

from pathlib import Path
spec = Path(spec)
assert(spec.exists())
spec = (spec.read_text())
from .program import Program
p = Program.parse(spec)
p.run()

exit(0)
